@@ -107,12 +107,136 @@ module Qt
 	# behaves just like a normal ruby String. TODO: override
 	# methods which alter the underlying string, in order to
 	# sync the QByteArray with the changed string.
+	# If the data arg is nil, the string is returned as the
+	# value instead. 
 	class ByteArray < DelegateClass(String)
+		attr_reader :private_data
 		attr_reader :data
 		
-		def initialize(string, data)
+		def initialize(string, data=nil)
 			super(string)
-			@data = data
+			@private_data = data
+			@data = string
+		end
+	end
+	
+	# Delete the underlying C++ instance after exec returns
+	# Otherwise, rb_gc_call_finalizer_at_exit() can delete
+	# stuff that Qt::Application still needs for its cleanup.
+	class Application < Qt::Base
+		def exec
+			super
+			self.dispose
+			Qt::Internal.application_terminated = true
+		end
+	end
+	
+	# Provides a mutable numeric class for passing to methods with
+	# C++ 'int*' or 'int&' arg types
+	class Integer
+		attr_accessor :value
+		def initialize(n=0) @value = n end
+		
+		def +(n) 
+			return Integer.new(@value + n.to_i) 
+		end
+		def -(n) 
+			return Integer.new(@value - n.to_i)
+		end
+		def *(n) 
+			return Integer.new(@value * n.to_i)
+		end
+		def /(n) 
+			return Integer.new(@value / n.to_i)
+		end
+		def %(n) 
+			return Integer.new(@value % n.to_i)
+		end
+		def **(n) 
+			return Integer.new(@value ** n.to_i)
+		end
+		
+		def |(n) 
+			return Integer.new(@value | n.to_i)
+		end
+		def &(n) 
+			return Integer.new(@value & n.to_i)
+		end
+		def ^(n) 
+			return Integer.new(@value ^ n.to_i)
+		end
+		def <<(n) 
+			return Integer.new(@value << n.to_i)
+		end
+		def >>(n) 
+			return Integer.new(@value >> n.to_i)
+		end
+		
+		def <=>(n)
+			if @value < n
+				return -1
+			elsif @value > n
+				return 1
+			else
+				return 0
+			end
+		end
+		
+		def to_f() return @value.to_f end
+		def to_i() return @value.to_i end
+		def to_s() return @value.to_s end
+	end
+	
+	# If a C++ enum was converted to an ordinary ruby Integer, the
+	# name of the type is lost. The enum type name is needed for overloaded
+	# method resolution when two methods differ only by an enum type.
+	class Enum < Qt::Integer
+		attr_accessor :type
+		def initialize(n, type)
+			super(n) 
+			@value = n 
+			@type = type
+		end
+		
+		def |(n) 
+			return Enum.new(@value | n.to_i, @type)
+		end
+		def &(n) 
+			return Enum.new(@value & n.to_i, @type)
+		end
+		def ^(n) 
+			return Enum.new(@value ^ n.to_i, @type)
+		end
+		def <(n) 
+			return @value < n.to_i
+		end
+		def <=(n) 
+			return @value <= n.to_i
+		end
+		def >(n) 
+			return @value > n.to_i
+		end
+		def >=(n) 
+			return @value >= n.to_i
+		end
+		def <<(n) 
+			return Enum.new(@value << n.to_i, @type)
+		end
+		def >>(n) 
+			return Enum.new(@value >> n.to_i, @type)
+		end
+		
+		def ==(n) return @value == n.to_i end
+		def to_i() return @value end
+	end
+	
+	# Provides a mutable boolean class for passing to methods with
+	# C++ 'bool*' or 'bool&' arg types
+	class Boolean
+		attr_accessor :value
+		def initialize(b=false) @value = b end
+		def nil? 
+			return !@value 
 		end
 	end
 	
@@ -152,20 +276,52 @@ module Qt
 		def checkarg(argtype, typename)
 			puts "      #{typename} (#{argtype})" if debug_level >= DebugLevel::High
 			if argtype == 'i'
-				if typename =~ /^int$/
+				if typename =~ /^int&?$|^signed$/
 					return 1
-				elsif typename =~ /^(?:short|ushort|uint|long|ulong|signed|unsigned)$/
+				elsif typename =~ /^(?:short|ushort|uint|long|ulong|unsigned|float|double)$/
 					return 0
+				else 
+					t = typename.sub(/^const\s+/, '')
+					t.sub!(/[&*]$/, '')
+					if isEnum(t)
+						return 0
+					end
 				end
 			elsif argtype == 'n'
-				if typename =~ /^(?:float|double)$/
+				if typename =~ /^double$/
+					return 2
+				elsif typename =~ /^float$/
+					return 1
+				elsif typename =~ /^int&?$/
+					return 0
+				elsif typename =~ /^(?:short|ushort|uint|long|ulong|signed|unsigned|float|double)$/
+					return 0
+				else 
+					t = typename.sub(/^const\s+/, '')
+					t.sub!(/[&*]$/, '')
+					if isEnum(t)
+						return 0
+					end
+				end
+			elsif argtype == 'B'
+				if typename =~ /^(?:bool)[*&]?$/
 					return 0
 				end
+			elsif argtype == 'b'
+				# An argtype 'b' means a Qt::ByteArray has been passed to a C++ method expecting a QByteArray arg.
+				# In that case a Qt::ByteArray must take precedence over any String, and scores 3. Note that a
+				# ruby String would only score 2 when passed to the same QByteArray method - an alternative 
+				# method expecting a QString arg would take precedence over it with a score of 3.
+				if typename =~ /^(const )?(QByteArray[*&]?)$/
+					return 3
+				end
 			elsif argtype == 's'
-				if typename =~ /^(?:u?char\*|const u?char\*|(?:const )?((Q(C?)String)|QByteArray)[*&]?)$/
+				if typename =~ /^(const )?((QByteArray|QChar)[*&]?)$/
+					return 1
+				elsif typename =~ /^(?:u?char\*|const u?char\*|(?:const )?(Q(C?)String)[*&]?)$/
 					qstring = !$1.nil?
-					c = ("C" == $3)
-					return c ? 1 : (qstring ? 2 : 0)
+					c = ("C" == $2)
+					return c ? 2 : (qstring ? 3 : 0)
 				end
 			elsif argtype == 'a'
 				# FIXME: shouldn't be hardcoded. Installed handlers should tell what ruby type they expect.
@@ -196,10 +352,15 @@ module Qt
 			else
 				t = typename.sub(/^const\s+/, '')
 				t.sub!(/[&*]$/, '')
-				if classEquals(argtype, t)
+				if argtype == t
 					return 1
 				elsif classIsa(argtype, t)
 					return 0
+				elsif isEnum(argtype) and 
+						(t =~ /int|uint|long|ulong|WFlags|WState|ProcessEventsFlags|ComparisonFlags/ or
+						t =~ /SFlags|SCFlags|WId|difference_type|ToolBarDock/ or
+						t =~ /KStyleFlags|KonqPopupFlags/)
+					return 0
 				end
 			end
 			return -99
@@ -251,7 +412,7 @@ module Qt
 			end
 			if method == "new"
 				method = classname.dup 
-				method.gsub!(/^(KParts|KIO|khtml|DOM)::/,"")
+				method.gsub!(/^(KParts|KIO|khtml|DOM|Kontact|Kate|KTextEditor)::/,"")
 			end
 			method = "operator" + method.sub("@","") if method !~ /[a-zA-Z]+/
 			# Change foobar= to setFoobar()					
@@ -281,6 +442,15 @@ module Qt
 			
 			methodIds = []
 			methods.collect { |meth| methodIds.concat( findMethod(classname, meth) ) }
+			
+			if method =~ /_/ && methodIds.length == 0
+				# If the method name contains underscores, convert to camel case
+				# form and try again
+				while method =~ /([^_]*)_(.)(.*)/ 
+					method = $1 + $2.upcase + $3
+				end
+				do_method_missing(package, method, klass, this, *args)
+			end
 
 			if debug_level >= DebugLevel::High
 				puts "classname    == #{classname}"
@@ -295,9 +465,7 @@ module Qt
 			end
 			
 			chosen = nil
-			if methodIds.length == 1 && method !~ /^operator/
-				chosen = methodIds[0]
-			elsif methodIds.length > 0
+			if methodIds.length > 0
 				best_match = -1
 				methodIds.each do
 					|id|
@@ -357,14 +525,44 @@ module Qt
 			# Special case QByteArray, as it's disguised as a ruby String
 			# and not in the public api.
 			@@classes['Qt::ByteArray'] = Qt::ByteArray.class
+			@@classes['Qt::Integer'] = Qt::Integer.class
+			@@classes['Qt::Boolean'] = Qt::Boolean.class
+			@@classes['Qt::Enum'] = Qt::Enum.class
 		end
 		
 		def create_qbytearray(string, data)
 			return Qt::ByteArray.new(string, data)
 		end
 		
-		def get_qbytearray(string)
-			return string.data
+		def get_qbytearray(str)
+			if str.private_data.nil?
+				return str.data
+			end
+			return str.private_data
+		end
+		
+		def get_qinteger(num)
+			return num.value
+		end
+		
+		def set_qinteger(num, val)
+			return num.value = val
+		end
+		
+		def create_qenum(num, type)
+			return Qt::Enum.new(num, type)
+		end
+		
+		def get_qenum_type(e)
+			return e.type
+		end
+		
+		def get_qboolean(b)
+			return b.value
+		end
+		
+		def set_qboolean(b, val)
+			return b.value = val
 		end
 	end
 
@@ -391,8 +589,11 @@ module Qt
 		
 		def add_signals(signal_list)
 			signal_list.each do |signal|
+				signal = Qt::Object.normalizeSignalSlot(signal)
 				if signal =~ /([^\s]*)\((.*)\)/
 					@signals.push QObjectMember.new($1, signal, $2)
+				else
+					qWarning( "#{@klass.name}: Invalid signal format: '#{signal}'" )
 				end
 			end
 		end
@@ -413,8 +614,11 @@ module Qt
 		
 		def add_slots(slot_list)
 			slot_list.each do |slot|
+				slot = Qt::Object.normalizeSignalSlot(slot)
 				if slot =~ /([^\s]*)\((.*)\)/
 					@slots.push QObjectMember.new($1, slot, $2)
+				else
+					qWarning( "#{@klass.name}: Invalid slot format: '#{slot}'" )
 				end
 			end
 		end
@@ -473,13 +677,14 @@ module Qt
 	end
 
 	def getMocArguments(member)
-		argStr = member.sub(/.*\(/, '').sub!(/\)$/, '')
-		args = argStr.scan(/[^,]+/)
+		argStr = member.sub(/.*\(/, '').sub(/\)$/, '')
+		args = argStr.scan(/([^,]*<[^>]+>)|([^,]+)/)
 		mocargs = allocateMocArguments(args.length)
                 args.each_with_index {
                         |arg, i|
+			arg = arg.to_s
 			a = arg.sub(/^const\s+/, '')
-	    	        a = (a =~ /^(bool|int|double|char\*|QString)&?$/) ? $1 : 'ptr'
+			a = (a =~ /^(bool|int|double|char\*|QString)&?$/) ? $1 : 'ptr'
 			valid = setMocType(mocargs, i, arg, a)
                 }
 		result = []
@@ -562,6 +767,7 @@ class Object
 	# so remove it..
 	undef_method :display
 	undef_method :type
+	undef_method :exec
 	def SIGNAL(string) ; return "2" + string; end
 	def SLOT(string)   ; return "1" + string; end
 	def emit(signal)   ; end
