@@ -22,6 +22,7 @@ namespace Qt {
 	using System;
 	using System.Collections;
 	using System.Reflection;
+	using System.Text.RegularExpressions;
 	using System.Runtime.InteropServices;
 
 	public class QtSupport: IDisposable {
@@ -166,12 +167,16 @@ namespace Qt {
 
 		public static string SIGNAL (string signal)
 		{
-			return "2"+signal.Replace (" ", "");
+			return "2"+
+				Regex.Replace (signal, @"\s*\(\s*\)", "()");
+			
+
 		}
 
 		public static string SLOT (string slot)
 		{
-			return "1"+slot.Replace (" ", "");
+			return "1" + 
+				Regex.Replace (slot, @"\s*\(\s*\)", "()");
 		}
 
 		/************************** Object Tracking/Boxing and Disposal *********************************/
